__all__ = [
    'connect',
    'orchestrate',
]

from pyrasgo.rasgo import Rasgo
from pyrasgo.orchestration import RasgoOrchestration


def connect(api_key):
    return Rasgo(api_key=api_key)

def orchestrate(api_key):
    return RasgoOrchestration(api_key=api_key)


from typing import Optional
from pydantic import BaseModel

from pyrasgo.schemas.data_source import DataSourceBase


class FeatureSetBase(BaseModel):
    id: int


class FeatureSetCreate(BaseModel):
    snowflakeTable: str
    name: str
    dataSource: DataSourceBase
    granularity: Optional[str]


class FeatureSetUpdate(BaseModel):
    id: int
    snowflakeTable: Optional[str]
    name: Optional[str]
    dataSource: Optional[DataSourceBase]
    granularity: Optional[str]


class FeatureSet(FeatureSetBase):
    name: str
    granularity: Optional[str]



import pandas as pd


def build_schema(dataframe: pd.DataFrame, include_index=False) -> dict:
    from pandas.io.json import build_table_schema
    schema_list = build_table_schema(dataframe)
    if not include_index:
        return {column['name']: column
                for column in schema_list['fields'] if column['name'] != 'index'}
    return {column['name']: column
            for column in schema_list['fields']}



import os


class Member(object):

    def __init__(self, api_object):
        self._api_object = api_object

    def organization_id(self):
        return self._api_object["organizationId"]

    def snowflake_creds(self):
        org = self._api_object.get("organization", {})
        user = self._api_object.get("snowUsername", os.environ.get('SNOWFLAKE_USERNAME', None))
        return {
            "user": user,
            "password": self._api_object.get("snowPassword", os.environ.get('SNOWFLAKE_PASSWORD')),
            "account": org.get("account", os.environ.get("SNOWFLAKE_ACCOUNT")),
            "database": org.get("database", os.environ.get("SNOWFLAKE_DATABASE")),
            "schema": org.get("schema", os.environ.get("SNOWFLAKE_SCHEMA")),
            "warehouse": org.get("warehouse", os.environ.get("SNOWFLAKE_WAREHOUSE")),
            "role": self._api_object.get("role", None) or f"{user}_role"
        }



from types import SimpleNamespace


class Namespace(SimpleNamespace):
    """
    Extended SimpleNamespace to allow for recursive dictionaries
    """

    def __init__(self, **kwargs):
        """Create a SimpleNamespace recursively"""
        super().__init__()
        self.__dict__.update({k: self.__parse(v) for k, v in kwargs.items()})

    def __parse(self, struc):
        """Recurse into nested data structure to create leaf namespace objects"""
        if type(struc) is dict:
            return type(self)(**struc)
        if type(struc) in (list, tuple):
            return [self.__parse(i) for i in struc]
        return struc

    def __unparse(self, value):
        if isinstance(value, type(self)):
            return value.to_dict()
        if type(value) in (list, tuple):
            return [self.__unparse(i) for i in value]
        if value is None:
            return
        return value

    def to_dict(self):
        return {k: self.__unparse(v) or 'null' for k, v in self.__dict__.items()}



from pyrasgo.connection import Connection
from pyrasgo.monitoring import track_usage
from pyrasgo.rasgo import Rasgo

class RasgoOrchestration(Connection):

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.rasgo = Rasgo(api_key=self._api_key)

    @track_usage
    def simulate_orchestration(self, source_table: str, func):
        '''
        Run a python function against a source table
        
        param source_table: Snowflake table holding raw data
        param func: function containing feature transformation code (should be named generate_feature)

        return: Success or Failure message
        '''
        df = self.rasgo.get_source_table(table_name=source_table, record_limit=-1)
        dx = func(df)
        return f'Code successfully created dataframe with shape {dx.shape}'



from typing import Optional
from pydantic import BaseModel
from pyrasgo.schemas.dimensionality import Dimensionality
from pyrasgo.schemas.feature_set import v0


class ColumnCreate(BaseModel):
    name: str
    dataType: str
    featureSet: v0.FeatureSetBase
    dimensionality: Dimensionality


class ColumnUpdate(BaseModel):
    id: int
    name: Optional[str]
    dataType: Optional[str]
    featureSet: Optional[v0.FeatureSetBase]
    dimensionality: Optional[Dimensionality]


class Column(BaseModel):
    id: Optional[int]



from enum import Enum
from typing import Union


class Granularity(Enum):
    """
    Enum class to describe allowed types of granularities
    """
    # TODO: This exists to support validation prior to posting from the API.
    #       For ease of maintenance, this should be handled through API returns.
    ACTION = 'action'
    ACTOR = 'actor'
    AIRLINE = 'airline'
    AIRPORT = 'airport'
    ATBAT = 'atbat'
    AUTHOR = 'author'
    BUYER = 'buyer'
    CBG = 'CBG'
    CITY = 'city'
    CLAIM = 'claim'
    COMPANY = 'company'
    COUNTY = 'county'
    COUNTRY = 'country'
    CUSTOM = 'custom'
    CUSTOMER = 'customer'
    DAY = 'day'
    DMA = 'DMA'
    FIPS = 'FIPS'
    FLIGHT = 'flight'
    GAME = 'game'
    HOUR = 'hour'
    LATLONG = 'latlong'
    MAKE = 'make'
    MEMBER = 'member'
    MODEL = 'model'
    MONTH = 'month'
    MOVIE = 'movie'
    NAME = 'name'
    PAYER = 'payer'
    PLAY = 'play'
    PLAYER = 'player'
    PITCH = 'pitch'
    PRODUCT = 'product'
    PROVIDER = 'provider'
    PUBLISHER = 'publisher'
    QUARTER = 'quarter'
    SEASON = 'season'
    STATE = 'state'
    STORE = 'store'
    STUDIO = 'studio'
    SUBSCRIBER = 'subscriber'
    TEAM = 'team'
    TERM = 'term'
    TICKER = 'ticker'
    TITLE = 'title'
    USER = 'user'
    UNIT = 'unit'
    WEEK = 'week'
    YEAR = 'year'
    ZIPCODE = 'zipcode' 


class ModelTypes(Enum):
    """
    Enum class to describe allowed types of models
    """
    # TODO: This exists to support validation prior to posting from the API.
    #       For ease of maintenance, this should be handled through API returns.
    TIMESERIES = 'time series'


class ModelType(object):
    """
    Class to wrap the ModelTypes enum for additional functionality

    (Extending the enum class is not allowed, and subclassing is possible, but unnecessary)
    """

    def __init__(self, name: str):
        self._type = ModelTypes(name)
        # TODO: Lookup compatible when other model types and/or granularities are supported.
        self._compatible_granularities = [Granularity.DAY, Granularity.MONTH,
                                          Granularity.YEAR, Granularity.WEEK, Granularity.QUARTER]

    @property
    def name(self):
        return self._type.name

    @property
    def value(self):
        return self._type.value

    @property
    def compatible_granularities(self):
        return self._compatible_granularities

    def is_compatible(self, granularity: Union[str, Granularity]) -> bool:
        # If not enum, convert to enum first.
        try:
            granularity.name
        except AttributeError:
            granularity = Granularity(granularity)
        return granularity in self._compatible_granularities



from pydantic import BaseModel
from typing import Optional

from pyrasgo.schemas.organization import Organization


class DataSourceBase(BaseModel):
    id: int


class DataSourceCreate(BaseModel):
    name: str
    abbreviation: str
    organization: Organization


class DataSource(DataSourceBase):
    name: Optional[str]  # TODO: Should not be optional
    category: Optional[str] = None



from types import FunctionType

from collections import OrderedDict
import logging
import os
import yaml

from uncompyle6.semantics import pysource

from pyrasgo.schemas.feature_set import v1


def load_feature_set_from_yaml(*, file_name: str, directory: str = None) -> v1.FeatureSet:
    if directory is None:
        directory = os.getcwd()

    if os.path.splitext(f"{directory}/{file_name}")[1] in ['yaml', 'yml']:
        raise ValueError("Must provide valid yaml file")

    with open(f"{directory}/{file_name}") as _yaml:
        return v1.FeatureSet.parse_obj(yaml.load(_yaml, Loader=yaml.SafeLoader)[0])


def save_feature_set_to_yaml(feature_set: v1.FeatureSet, *,
                             file_name: str, directory: str = None, overwrite: bool = True) -> None:
    if directory is None:
        directory = os.getcwd()

    if directory[-1] == "/":
        directory = directory[:-1]

    if file_name.split(".")[-1] not in ['yaml', 'yml']:
        file_name += ".yaml"

    if os.path.exists(f"{directory}/{file_name}") and overwrite:
        logging.warning(f"Overwriting existing file {file_name} in directory: {directory}")

    safe_dumper = yaml.SafeDumper
    safe_dumper.add_representer(v1.DataType, lambda self, data: self.represent_str(str(data.value)))
    safe_dumper.add_representer(OrderedDict, lambda self, data: self.represent_mapping('tag:yaml.org,2002:map',
                                                                                       data.items()))
    safe_dumper.ignore_aliases = lambda self, data: True

    with open(f"{directory}/{file_name}", "w") as _yaml:
        # TODO: Not truly specifying an order yet.
        yaml.dump(data=OrderedDict(feature_set.dict(exclude_unset=True)), Dumper=safe_dumper, stream=_yaml)


def generate_feature_set_files(source_table: str, name: str = None, *,
                               feature_set: v1.FeatureSet, function: FunctionType = None,
                               directory: str = None, overwrite: bool = False) -> tuple:
    """
    return: yml file
    """
    if directory is None:
        directory = os.getcwd()
    working_directory = f"{directory}/featureSets/{name or source_table}"
    os.makedirs(working_directory, exist_ok=True)

    feature_set_yaml = f"{source_table}.yaml"
    requirements_txt = "requirements.txt"
    function_py = f"{function.__name__}.py" if function else ""
    files_to_create = [feature_set_yaml] if function_py == "" else [feature_set_yaml, requirements_txt, function_py]

    for existing_file in os.listdir(working_directory):
        if existing_file == requirements_txt and not overwrite:
            raise ValueError(f"There is {requirements_txt} existing in the directory {working_directory}")

        if existing_file in files_to_create:
            logging.warning(f"The directory {working_directory} contains the file "
                            f"({existing_file}) it will be overwritten.")

    if function:
        with open(f"{working_directory}/{function_py}", 'w') as _py:
            _py.write(get_function_code(function=function))

        with open(f"{working_directory}/{requirements_txt}", 'w') as _txt:
            _txt.writelines([f"{requirement.project_name}=={requirement.parsed_version}\n"
                             for requirement in identify_requirements(function=function)])

    save_feature_set_to_yaml(feature_set, file_name=feature_set_yaml, directory=working_directory)

    return feature_set_yaml, function_py, requirements_txt if function else feature_set_yaml, None, None


def get_function_code(*, function: FunctionType):
    """
    return: feature.py file
    """
    import_statements = []
    number_of_arguments = function.__code__.co_argcount
    arguments_clause = ",".join(function.__code__.co_varnames[0:number_of_arguments]) if number_of_arguments > 0 else ''
    indented = [""]
    def_signature = ["", f"def {function.__name__}({arguments_clause}):"]
    for line in pysource.code_deparse(function.__code__).text.split("\n"):
        if line.startswith("import"):
            import_statements.append(line)
        else:
            indented.append(f"{4 * ' '}{line}")
    return "\n".join(import_statements + def_signature + indented)


def identify_requirements(*, function: FunctionType):
    """
    return: content of requirements.txt to execute a function
    """
    imported_libraries = function.__code__.co_names
    try:
        from pip._internal.utils.misc import get_installed_distributions
    except ImportError:
        from pip import get_installed_distributions

    return [dist for dist in get_installed_distributions() if dist.project_name in imported_libraries]



from typing import List, Optional
from pydantic import BaseModel

from pyrasgo.schemas.data_source import DataSource
from pyrasgo.schemas.granularity import Granularity
from pyrasgo.schemas.feature_set import v0


class FeatureBase(BaseModel):
    id: int


class FeatureCreate(BaseModel):
    name: str
    code: Optional[str]
    description: Optional[str]
    columnId: Optional[int]
    featureSetId: Optional[int]
    organizationId: Optional[int]
    orchestrationStatus: Optional[str]
    tags: Optional[List[str]]

class FeatureUpdate(BaseModel):
    id: int
    name: Optional[str]
    code: Optional[str]
    description: Optional[str]
    #Can't include these until we create a v1 patch endpoint for features
    #columnId: Optional[int]
    #featureSetId: Optional[int]
    #organizationId: Optional[int]
    #orchestrationStatus: Optional[str]
    #tags: Optional[List[str]]

class Feature(FeatureBase):
    id: str  # TODO: Returning feature ids as strings, ensure consistency.
    name: str
    description: Optional[str]
    code: Optional[str]
    columnId: Optional[int]
    dataType: Optional[str]
    dataSource: Optional[DataSource]
    featureSet: Optional[v0.FeatureSet]
    granularities: Optional[List[Granularity]]
    orchestrationStatus: Optional[str]
    tags: Optional[List[str]]

class FeatureStats(BaseModel):
    recCt: int
    distinctCt: int
    nullRecCt: int
    zeroValRecCt: int
    meanVal: str
    medianVal: str
    maxVal: str
    minVal: str
    sumVal: str
    stdDevVal: str
    varianceVal: str
    rangeVal: str
    skewVal: str
    kurtosisVal: str
    q1Val: str
    q3Val: str
    IQRVal: str
    pct5Val: str
    pct95Val: str